# Plugin for TWiki Collaboration Platform, http://TWiki.org/
#
# Copyright (C) 2000-2003 Andrea Sterbini, a.sterbini@flashnet.it
# Copyright (C) 2001-2003 Peter Thoeny, peter@thoeny.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details, published at 
# http://www.gnu.org/copyleft/gpl.html
# =========================
package TWiki::Plugins::AliasPlugin;    # change the package name and $pluginName!!!
# =========================
use vars qw(
        $web $topic $user $installWeb $VERSION
        $aliasPattern $debug $acceptNonTWikiWord
	$aliasWeb $aliasTopic $wordRegex $wikiWordRegex
	$defaultAliasTopic
    );

$VERSION = '1.010';

# =========================
sub writeDebug {
  TWiki::Func::writeDebug("AliasPlugin - " . $_[0]) if $debug;
}

# =========================
sub initPlugin
{
  ( $topic, $web, $user, $installWeb ) = @_;

  # check for Plugins.pm versions
  if ($TWiki::Plugins::VERSION < 1) {
    TWiki::Func::writeWarning( "Version mismatch between AliasPlugin and Plugins.pm" );
    return 0;
  }

  # 0: off, 1: debug, 2: heavy debug
  $debug = 0;

  # for getRegularExpression
  if ($TWiki::Plugins::VERSION < 1.020) {
    eval 'use TWiki::Contrib::CairoContrib;';
    #writeDebug("reading in CairoContrib");
  }


  # get plugin flags
  $acceptNonTWikiWord = 
    TWiki::Func::getPreferencesFlag("ALIASPLUGIN_ACCEPT_NON_TWIKI_WORD_ALIASES");
  $defaultAliasTopic = 
    TWiki::Func::getPreferencesValue("ALIASPLUGIN_DEFAULT_ALIASES");
  
  # decide on how to match alias words
  $wikiWordRegex = &TWiki::Func::getRegularExpression('wikiWordRegex');
  if ($acceptNonTWikiWord) {
    $wordRegex = '\w+';
  } else {
    $wordRegex = $wikiWordRegex;
  }

  # set the default alias definitions
  $aliasWeb = '';
  $aliasTopic = '';
  &getAliases(0);

  # Plugin correctly initialized
  writeDebug("initPlugin( $web.$topic ) is OK" );

  return 1;
}

# =========================
sub commonTagsHandler
{
  #my ($text, $topic, $web) = @_; dont uncomment
  
  #&writeDebug ("commonTagsHandler()");
  $_[0] =~ s/%ALIASES{(.*?)}%/&handleAliases($1)/ge;
  $_[0] =~ s/%ALIAS{(.*?)}%/&handleAlias($1)/ge;
  $_[0] =~ s/%UNALIAS{(.*?)}%/&handleUnAlias($1)/ge;

  $_[0] =~ s/%UNALIAS%/&handleUnAlias()/ge;
  $_[0] =~ s/%ALIASES%/&handleAliases()/ge;
}

# =========================
sub handleAliases {
  my $args = shift;

  writeDebug("handleAliases() called");

  if ($args) {
    my $doMerge = &TWiki::Func::extractNameValuePair($args, "merge") || 'off';
    if ($doMerge eq 'on') {
      $doMerge = 1;
    } elsif ($doMerge eq 'off') {
      $doMerge = 0;
    } else {
      return '<font color="red">' .
	     'Error in %<nop>ALIASES%: =merge= must be =on= or =off= </font>';
    }

    return "" if &getAliases($doMerge, &TWiki::Func::extractNameValuePair($args));
    return '<font color="red">' .
	   'Error in %<nop>ALIASES%: no alias definitions found</font>';
  }

  my $text = "<noautolink>\n";
  $text .= "| *Name* | *Value* |\n";
  foreach my $key (sort {uc($a) cmp uc($b)} keys %aliasHash) {
    $text .= "| <nop>$key | $aliasHash{$key} |\n";
  }
  $text .= "</noautolink>\n";
  
  return $text;
}

# =========================
sub handleAlias {
  my $args = shift;

  writeDebug("handleAlias() called");

  my $thisKey = &TWiki::Func::extractNameValuePair($args, 'name');
  my $thisValue = &TWiki::Func::extractNameValuePair($args, 'value');

  if ($thisKey && $thisValue) {
    $aliasHash{$thisKey} = $thisValue;
    $aliasPattern = join('|', keys(%aliasHash));
    writeDebug("handleAlias(): added alias '$thisKey' -> '$thisValue')");
    writeDebug("handleAlias(): aliasPattern='$aliasPattern'") if $aliasPattern;
    return "";
  }

  return '<font color="red">Error in %<nop>ALIAS%: need a =name= and a =value= </font>';
}

# =========================
sub handleUnAlias {
  my $args = shift;

  writeDebug("handleUnAlias() called");

  if ($args) {
    my $thisKey = &TWiki::Func::extractNameValuePair($args) ||
		  &TWiki::Func::extractNameValuePair($args, 'name');

    if ($thisKey) {
      my $thisValue = $aliasHash{$thisKey};
      if ($thisValue) {
	delete $aliasHash{$thisKey};
	$aliasPattern = join('|', keys(%aliasHash));
	writeDebug("handleUnAlias(): removed alias '$thisKey' -> '$thisValue')");
      }
      return "";
    }

    return '<font color="red">Error in %<nop>UNALIAS%: don\'t know what to unalias</font>';
  } 

  writeDebug("handleUnAlias(): dumping all aliases");
  %aliasHash = ();
}


# =========================
sub outsidePREHandler
{
  return unless $aliasPattern;

  my ($text) = @_;

  #writeDebug("outsidePREHandler() - before: '$text'") if $debug > 1;
  $text =~ s/<nop>/NOPTOKEN/g;
  {
    my $result = '';
    # escape html tags
    while ($text =~ /([^<]*)((?:<[^>]*>)*|<)/g) {
      my $substr = $1;
      my $tail = $2;
      if ($substr) {
        # escape twiki tags
	while ($substr =~ /([^%]*)(((%[A-Z][a-zA-Z_0-9]+({[^}]+})?%)*)|%)/g) {
	  my $substr = $1;
	  my $tail = $2;
	  if ($substr) {
	    # escape twiki links
	    while ($substr =~ /([^\[]*)((?:\[[^\]]*\])*|\[)/g) {
	      my $substr = $1;
	      my $tail = $2;
	      if ($substr) {
		# do the substitution
		if ($debug > 1) { 
		  $substr =~ s/\b($aliasPattern)\b/&_debugSubstitute($1)/ge;
		} else {
		  $substr =~ s/\b($aliasPattern)\b/$aliasHash{$1}/g;
		}
		$result .= $substr;
	      }
	      $result .= $tail if $tail;
	    }
	  }
	  $result .= $tail if $tail;
	}
      }
      $result .= $tail if $tail;
    }
    $text = $result;
  }
  $text =~ s/NOPTOKEN/<nop>/g;
  #writeDebug("outsidePREHandler() - after:  '$text'") if $debug > 1;

  $_[0] = $text
}

# =========================
sub _debugSubstitute
{
  my ($key) = @_;
  my $value = $aliasHash{$key};
  writeDebug("subst '$key' -> '$value'");
  return $value; 
}

# =========================
sub getAliases
{
  my ($doMerge, $thisWebTopic) = @_;
  my $thisWeb;
  my $thisTopic;


  # extract web and topic name
  my $topicRegex = &TWiki::Func::getRegularExpression('mixedAlphaNumRegex');
  my $webRegex = &TWiki::Func::getRegularExpression('webNameRegex');

  $thisWebTopic = $defaultAliasTopic unless $thisWebTopic;
  $thisWebTopic =~ s/^\s+//o;
  $thisWebTopic =~ s/\s+$//o;
  if ($thisWebTopic =~ /^($topicRegex)$/) {
    $thisTopic = $1;
    $thisWeb = $web;
  } elsif ($thisWebTopic =~ /^($webRegex)\.($topicRegex)$/) {
    $thisWeb = $1;
    $thisTopic = $2;
  }

  # have we alread red these aliaes
  if ($thisTopic eq $aliasTopic && $thisWeb eq $aliasWeb) {
    writeDebug("bailing out on $thisWeb.$thisTopic");
    return
  }

  # find topic with alias definitions
  $aliasWeb = $thisWeb;
  $aliasTopic = $thisTopic;

  # look for thisWeb.thisTopic
  writeDebug("looking for $thisWeb.$thisTopic");
  if (!&TWiki::Func::topicExists($thisWeb, $thisTopic)) {

    # look for TWIKIWEB.thisTopic
    $thisWeb = &TWiki::Func::getTwikiWebname();
    writeDebug("looking for $thisWeb.$thisTopic");
    if (!&TWiki::Func::topicExists($thisWeb, $thisTopic)) {

      # nothing there
      writeDebug("getAliases($webTopic) - no alias definitions found");
      return 0;
    }
  }
  $aliasWeb = $thisWeb;
  $aliasTopic = $thisTopic;
  writeDebug("reading aliases from $aliasWeb.$aliasTopic");

  # parse the plugin preferences lines
  my $prefText = TWiki::Func::readTopicText($aliasWeb, $aliasTopic);

  #$prefText =~ s/%INCLUDE{(.*?)}%/&handleIncludeFile($1, $topic, $web)/ge;

  if (!$doMerge) {
    writeDebug("getAliases(): dumping old aliases");
    %aliasHash = ();
  } else {
    writeDebug("getAliases(): merging aliases");
  }

  foreach my $line (split /\n/, $prefText) {
    if ($line =~ /^(?:\t| {3})+\* (?:\<nop\>)?($wordRegex): +(.*)$/) {
      my $key = $1;
      my $value = $2;
      $value =~ s/\s+$//go;

      # convenience for alias-links
      if ($value =~ /([^.]+)\.$wikiWordRegex/) {
	$value = "\[\[$value\]\[$key\]\]";
      }

      $aliasHash{$key} = $value;
      writeDebug("config match: key='$key', value='$value'");
    }
  }

  # create alias pattern
  $aliasPattern = join('|', keys(%aliasHash));
  writeDebug("getAliaes(): aliasPattern='$aliasPattern'") if $aliasPattern;

  return 1;
}

1;
